/*
 * refit/scan/securebootkeys.h
 *
 * Copyright (c) 2006-2010 Christoph Pfisterer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *  * Neither the name of Christoph Pfisterer nor the names of the
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "CloverDB.h"
#include "CanonicalDB.h"
#include "MSPCADB.h"
#include "MSUEFICADB.h"
#include "CloverKEK.h"
#include "MSKEK.h"

// Secure boot platform key
STATIC CONST UINT8 gSecureBootPlatformSignedKey[] = {
  0xde, 0x07, 0x0b, 0x05, 0x01, 0x37, 0x22, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xc0, 0x05, 0x00, 0x00, 0x00, 0x02, 0xf1, 0x0e, 0x9d, 0xd2, 0xaf, 0x4a, 0xdf, 0x68, 0xee, 0x49,
  0x8a, 0xa9, 0x34, 0x7d, 0x37, 0x56, 0x65, 0xa7, 0x30, 0x82, 0x05, 0xa4, 0x06, 0x09, 0x2a, 0x86,
  0x48, 0x86, 0xf7, 0x0d, 0x01, 0x07, 0x02, 0xa0, 0x82, 0x05, 0x95, 0x30, 0x82, 0x05, 0x91, 0x02,
  0x01, 0x01, 0x31, 0x0f, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02,
  0x01, 0x05, 0x00, 0x30, 0x0b, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x07, 0x01,
  0xa0, 0x82, 0x03, 0x23, 0x30, 0x82, 0x03, 0x1f, 0x30, 0x82, 0x02, 0x07, 0xa0, 0x03, 0x02, 0x01,
  0x02, 0x02, 0x09, 0x00, 0xf0, 0x33, 0x6a, 0xed, 0x40, 0x72, 0xc5, 0x15, 0x30, 0x0d, 0x06, 0x09,
  0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05, 0x00, 0x30, 0x26, 0x31, 0x24, 0x30,
  0x22, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x1b, 0x43, 0x6c, 0x6f, 0x76, 0x65, 0x72, 0x20, 0x50,
  0x6c, 0x61, 0x74, 0x66, 0x6f, 0x72, 0x6d, 0x20, 0x43, 0x65, 0x72, 0x74, 0x69, 0x66, 0x69, 0x63,
  0x61, 0x74, 0x65, 0x30, 0x1e, 0x17, 0x0d, 0x31, 0x33, 0x31, 0x32, 0x30, 0x35, 0x30, 0x31, 0x34,
  0x36, 0x31, 0x37, 0x5a, 0x17, 0x0d, 0x32, 0x33, 0x31, 0x32, 0x30, 0x33, 0x30, 0x31, 0x34, 0x36,
  0x31, 0x37, 0x5a, 0x30, 0x26, 0x31, 0x24, 0x30, 0x22, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x1b,
  0x43, 0x6c, 0x6f, 0x76, 0x65, 0x72, 0x20, 0x50, 0x6c, 0x61, 0x74, 0x66, 0x6f, 0x72, 0x6d, 0x20,
  0x43, 0x65, 0x72, 0x74, 0x69, 0x66, 0x69, 0x63, 0x61, 0x74, 0x65, 0x30, 0x82, 0x01, 0x22, 0x30,
  0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x82,
  0x01, 0x0f, 0x00, 0x30, 0x82, 0x01, 0x0a, 0x02, 0x82, 0x01, 0x01, 0x00, 0xcb, 0xb1, 0x82, 0x68,
  0xc1, 0x71, 0xab, 0xf5, 0x35, 0x03, 0xaa, 0x0e, 0xf7, 0x2e, 0x6d, 0x4b, 0x63, 0x98, 0x8f, 0xd6,
  0xae, 0xb0, 0xeb, 0x95, 0x36, 0x70, 0xf9, 0x9d, 0xe4, 0xf5, 0x6b, 0x0d, 0xc0, 0x81, 0x36, 0x1d,
  0xcf, 0x01, 0x45, 0x4c, 0xd6, 0xf0, 0xc5, 0x4e, 0x7b, 0x4e, 0x49, 0x1a, 0x84, 0x64, 0xc8, 0x7c,
  0x87, 0xb8, 0xf3, 0xb2, 0x0a, 0xa1, 0x1e, 0xd6, 0xb7, 0x42, 0x2d, 0xc4, 0x79, 0x58, 0x19, 0x78,
  0x9c, 0x7e, 0x76, 0x8d, 0xfa, 0xb9, 0x6a, 0xdc, 0x5b, 0xf8, 0x01, 0xa9, 0x53, 0x71, 0xc7, 0x0a,
  0x9d, 0x73, 0x2c, 0x4c, 0xdf, 0xe5, 0x88, 0xb4, 0x10, 0x55, 0xe7, 0xea, 0xce, 0xca, 0xc8, 0xdd,
  0x8a, 0xdb, 0x93, 0x43, 0x14, 0x59, 0xb4, 0x04, 0xab, 0xc7, 0xdc, 0xc7, 0x77, 0xc4, 0x85, 0xf6,
  0x02, 0x59, 0x22, 0x91, 0xd0, 0x32, 0x78, 0x47, 0xf2, 0x2d, 0x0d, 0xd4, 0x81, 0x95, 0x8f, 0x8d,
  0xd0, 0x8d, 0x35, 0x1f, 0x7c, 0x2f, 0x44, 0x13, 0x2b, 0x65, 0x3e, 0x78, 0xba, 0x02, 0x35, 0x2b,
  0x56, 0x82, 0x3f, 0x44, 0x56, 0x6b, 0x78, 0x8a, 0x6e, 0xd7, 0xa2, 0xd2, 0xfd, 0xcc, 0x85, 0x43,
  0x8d, 0x73, 0xd8, 0xb4, 0x2f, 0x60, 0xa9, 0x04, 0xc2, 0xa4, 0xc1, 0x8c, 0xfa, 0x70, 0x0f, 0xe6,
  0xdb, 0xcf, 0xbe, 0x95, 0xd2, 0xce, 0xa6, 0x3a, 0x62, 0xed, 0x6d, 0xc8, 0xa4, 0xa1, 0x8a, 0xe2,
  0xb9, 0x01, 0xfd, 0x39, 0x81, 0xa7, 0x2d, 0x6c, 0xa0, 0x8d, 0xc2, 0x49, 0xdb, 0x3a, 0xc7, 0x46,
  0x9b, 0x46, 0x08, 0x17, 0xd4, 0xbb, 0x78, 0xb6, 0x36, 0x5c, 0xc8, 0x79, 0x9f, 0x62, 0xe7, 0xe2,
  0x46, 0x07, 0x32, 0xbc, 0x18, 0xad, 0xf0, 0xb5, 0x24, 0x34, 0x49, 0xfd, 0xe0, 0x85, 0xcb, 0x22,
  0xaa, 0xbc, 0x3a, 0x01, 0x36, 0x1a, 0x1a, 0x07, 0x05, 0x90, 0xb1, 0xf5, 0x02, 0x03, 0x01, 0x00,
  0x01, 0xa3, 0x50, 0x30, 0x4e, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x1d, 0x0e, 0x04, 0x16, 0x04, 0x14,
  0xb4, 0xc6, 0xdc, 0xeb, 0x96, 0x35, 0xe5, 0xe4, 0xde, 0xf9, 0x53, 0x75, 0x62, 0xf7, 0x7c, 0x79,
  0x1a, 0x45, 0x91, 0x3e, 0x30, 0x1f, 0x06, 0x03, 0x55, 0x1d, 0x23, 0x04, 0x18, 0x30, 0x16, 0x80,
  0x14, 0xb4, 0xc6, 0xdc, 0xeb, 0x96, 0x35, 0xe5, 0xe4, 0xde, 0xf9, 0x53, 0x75, 0x62, 0xf7, 0x7c,
  0x79, 0x1a, 0x45, 0x91, 0x3e, 0x30, 0x0c, 0x06, 0x03, 0x55, 0x1d, 0x13, 0x04, 0x05, 0x30, 0x03,
  0x01, 0x01, 0xff, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b,
  0x05, 0x00, 0x03, 0x82, 0x01, 0x01, 0x00, 0x05, 0xda, 0x57, 0xea, 0x10, 0xfd, 0xe6, 0x19, 0x6c,
  0x37, 0xde, 0xac, 0x5a, 0xe0, 0x6a, 0x87, 0x17, 0x94, 0x80, 0x5e, 0xa6, 0xe8, 0xb8, 0x8c, 0xe6,
  0x43, 0xcc, 0x8b, 0x1d, 0x96, 0x87, 0xf2, 0xd9, 0xfd, 0xc7, 0xb3, 0x1c, 0x67, 0x7e, 0xf0, 0x60,
  0x5e, 0x2b, 0xdd, 0x60, 0xc7, 0x18, 0xcc, 0x34, 0x8e, 0x0b, 0x4e, 0xda, 0x80, 0x30, 0x4b, 0xc9,
  0x66, 0xeb, 0x3d, 0x95, 0xf6, 0xac, 0xd8, 0x76, 0x7e, 0x40, 0x30, 0x6f, 0xb0, 0xb3, 0xf6, 0x7d,
  0x13, 0x99, 0x69, 0xdc, 0x6a, 0xd4, 0xd1, 0xc0, 0x71, 0x69, 0xf3, 0x14, 0x9c, 0x9e, 0xe1, 0xb9,
  0x8e, 0x1e, 0xcb, 0x6b, 0x4e, 0xc9, 0xc0, 0x4e, 0x22, 0x4a, 0xe7, 0xd2, 0x37, 0x33, 0xd0, 0x15,
  0xb9, 0x24, 0x04, 0xe8, 0xba, 0x6c, 0x25, 0x29, 0x67, 0xd1, 0xfb, 0x16, 0x11, 0xca, 0x68, 0xc4,
  0x7c, 0x3f, 0x42, 0x6f, 0x7c, 0x59, 0x03, 0xf0, 0x59, 0x59, 0xd4, 0x72, 0x61, 0x8c, 0xb7, 0xc4,
  0xe8, 0xf0, 0x4b, 0x98, 0x45, 0x95, 0x09, 0x77, 0x17, 0x94, 0x9f, 0xcc, 0xe4, 0xaa, 0x9d, 0xf0,
  0x8f, 0xd8, 0x1c, 0x20, 0x2a, 0x92, 0xe0, 0x74, 0x0d, 0x05, 0xf0, 0xe7, 0x4f, 0x25, 0xd4, 0x72,
  0xa1, 0xa4, 0xf1, 0x9a, 0x37, 0x84, 0x1f, 0x75, 0xb0, 0x8f, 0xc9, 0x71, 0x3e, 0x25, 0x28, 0x96,
  0xf6, 0xd8, 0x40, 0xa3, 0x4e, 0xd7, 0xaa, 0x62, 0xe5, 0x87, 0xac, 0x06, 0x3f, 0x3c, 0x04, 0xe6,
  0x64, 0x74, 0x37, 0x05, 0xe6, 0x65, 0xaa, 0xc0, 0x8f, 0x6c, 0xb7, 0x2f, 0x93, 0xf9, 0x95, 0x6a,
  0x9d, 0x3c, 0xc7, 0xbb, 0x20, 0x86, 0x3e, 0x05, 0x04, 0x2a, 0xc3, 0xe1, 0xcb, 0x5f, 0x32, 0xd7,
  0xa3, 0x9f, 0xb3, 0x1d, 0x84, 0xeb, 0x30, 0xeb, 0x8a, 0x10, 0xf0, 0xc7, 0x38, 0x2c, 0x18, 0x74,
  0x2f, 0x76, 0x0a, 0x01, 0xda, 0xa5, 0x4f, 0x31, 0x82, 0x02, 0x45, 0x30, 0x82, 0x02, 0x41, 0x02,
  0x01, 0x01, 0x30, 0x33, 0x30, 0x26, 0x31, 0x24, 0x30, 0x22, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c,
  0x1b, 0x43, 0x6c, 0x6f, 0x76, 0x65, 0x72, 0x20, 0x50, 0x6c, 0x61, 0x74, 0x66, 0x6f, 0x72, 0x6d,
  0x20, 0x43, 0x65, 0x72, 0x74, 0x69, 0x66, 0x69, 0x63, 0x61, 0x74, 0x65, 0x02, 0x09, 0x00, 0xf0,
  0x33, 0x6a, 0xed, 0x40, 0x72, 0xc5, 0x15, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65,
  0x03, 0x04, 0x02, 0x01, 0x05, 0x00, 0xa0, 0x81, 0xe4, 0x30, 0x18, 0x06, 0x09, 0x2a, 0x86, 0x48,
  0x86, 0xf7, 0x0d, 0x01, 0x09, 0x03, 0x31, 0x0b, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
  0x01, 0x07, 0x01, 0x30, 0x1c, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x09, 0x05,
  0x31, 0x0f, 0x17, 0x0d, 0x31, 0x33, 0x31, 0x32, 0x30, 0x35, 0x30, 0x31, 0x35, 0x35, 0x33, 0x34,
  0x5a, 0x30, 0x2f, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x09, 0x04, 0x31, 0x22,
  0x04, 0x20, 0x9f, 0xac, 0xf0, 0x58, 0x65, 0x7c, 0x0d, 0xd6, 0xf5, 0xc5, 0x87, 0xd8, 0x7c, 0xc1,
  0x29, 0xde, 0x6e, 0xdc, 0x81, 0x7c, 0x5c, 0x64, 0xcf, 0xd5, 0xf0, 0xab, 0x62, 0xa5, 0x18, 0x94,
  0xa7, 0xc3, 0x30, 0x79, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x09, 0x0f, 0x31,
  0x6c, 0x30, 0x6a, 0x30, 0x0b, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x01, 0x2a,
  0x30, 0x0b, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x01, 0x16, 0x30, 0x0b, 0x06,
  0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x01, 0x02, 0x30, 0x0a, 0x06, 0x08, 0x2a, 0x86,
  0x48, 0x86, 0xf7, 0x0d, 0x03, 0x07, 0x30, 0x0e, 0x06, 0x08, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
  0x03, 0x02, 0x02, 0x02, 0x00, 0x80, 0x30, 0x0d, 0x06, 0x08, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
  0x03, 0x02, 0x02, 0x01, 0x40, 0x30, 0x07, 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 0x07, 0x30, 0x0d,
  0x06, 0x08, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x03, 0x02, 0x02, 0x01, 0x28, 0x30, 0x0d, 0x06,
  0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x04, 0x82, 0x01, 0x00,
  0x06, 0x80, 0x28, 0x6a, 0x64, 0x77, 0x33, 0x07, 0x55, 0xa6, 0x46, 0x6e, 0x85, 0x28, 0x57, 0xa1,
  0xa7, 0x1e, 0x38, 0xbb, 0xa2, 0x6e, 0x71, 0x8a, 0x33, 0x09, 0xc3, 0x01, 0x21, 0xfb, 0xc2, 0x23,
  0x48, 0x27, 0x86, 0x4a, 0x87, 0xd8, 0x2f, 0x6b, 0x77, 0xe3, 0xd2, 0xbf, 0x75, 0x5c, 0xd2, 0x47,
  0xf3, 0xbf, 0x0a, 0xbe, 0xca, 0x41, 0x65, 0x90, 0xfb, 0x5b, 0x0c, 0xaf, 0x3f, 0x9a, 0x84, 0x9d,
  0x88, 0x40, 0x96, 0x78, 0x06, 0x20, 0x12, 0xd2, 0xce, 0xf6, 0x96, 0xcb, 0xb6, 0xcb, 0xdb, 0xb6,
  0xb7, 0x2b, 0xc3, 0x30, 0xea, 0xd6, 0x4e, 0xbb, 0x2e, 0x81, 0x0b, 0x57, 0x9b, 0x97, 0xd1, 0x85,
  0xed, 0x10, 0xa2, 0x27, 0x89, 0xe1, 0x66, 0xab, 0x43, 0x6b, 0x5d, 0xeb, 0x9a, 0x65, 0x6d, 0x36,
  0x1b, 0xe1, 0xe3, 0x46, 0xba, 0xaa, 0xf3, 0x8c, 0xc8, 0x91, 0x5d, 0x04, 0xc7, 0xb9, 0x27, 0xd7,
  0xae, 0x0b, 0x05, 0x97, 0x8b, 0x90, 0x57, 0x63, 0xf4, 0xb0, 0x2d, 0x9a, 0x3d, 0x64, 0x44, 0xfb,
  0x45, 0x86, 0x90, 0xed, 0x54, 0x6c, 0x93, 0x84, 0xf0, 0x09, 0xdf, 0xdc, 0x72, 0xb7, 0x21, 0xa2,
  0x66, 0xfa, 0x06, 0x6e, 0xbb, 0x97, 0x07, 0x5b, 0x99, 0xd9, 0x60, 0x77, 0x0b, 0xaf, 0xe6, 0x8f,
  0x66, 0x0b, 0x09, 0x44, 0xae, 0x82, 0x6e, 0xcc, 0x66, 0x81, 0x21, 0x2a, 0x81, 0xd6, 0x5c, 0xf2,
  0x53, 0x23, 0x4c, 0x3b, 0x68, 0x39, 0xf7, 0x00, 0x69, 0xaf, 0xa3, 0x09, 0x6f, 0x40, 0x71, 0xd9,
  0x4f, 0x5e, 0x52, 0x41, 0x31, 0x5f, 0xab, 0xe5, 0x62, 0x8b, 0x32, 0x49, 0xdf, 0x5d, 0x8c, 0x9d,
  0xea, 0x68, 0x8c, 0x08, 0x3c, 0xc9, 0x99, 0x77, 0x4e, 0xa8, 0xa4, 0xc9, 0x8a, 0xdc, 0x88, 0x98,
  0xac, 0xcb, 0x13, 0x03, 0x59, 0x63, 0xb5, 0xcf, 0xb2, 0xe0, 0x3e, 0x10, 0x4f, 0x6c, 0x0d, 0x7d,
  0xa1, 0x59, 0xc0, 0xa5, 0xe4, 0x94, 0xa7, 0x4a, 0x87, 0xb5, 0xab, 0x15, 0x5c, 0x2b, 0xf0, 0x72,
  0x4f, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x33, 0x03, 0x00, 0x00, 0xe3, 0x28, 0x41, 0x5c,
  0x03, 0xb5, 0x14, 0x45, 0xbe, 0x93, 0x01, 0x6e, 0xab, 0x3f, 0x99, 0x11, 0x30, 0x82, 0x03, 0x1f,
  0x30, 0x82, 0x02, 0x07, 0xa0, 0x03, 0x02, 0x01, 0x02, 0x02, 0x09, 0x00, 0xf0, 0x33, 0x6a, 0xed,
  0x40, 0x72, 0xc5, 0x15, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01,
  0x0b, 0x05, 0x00, 0x30, 0x26, 0x31, 0x24, 0x30, 0x22, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x1b,
  0x43, 0x6c, 0x6f, 0x76, 0x65, 0x72, 0x20, 0x50, 0x6c, 0x61, 0x74, 0x66, 0x6f, 0x72, 0x6d, 0x20,
  0x43, 0x65, 0x72, 0x74, 0x69, 0x66, 0x69, 0x63, 0x61, 0x74, 0x65, 0x30, 0x1e, 0x17, 0x0d, 0x31,
  0x33, 0x31, 0x32, 0x30, 0x35, 0x30, 0x31, 0x34, 0x36, 0x31, 0x37, 0x5a, 0x17, 0x0d, 0x32, 0x33,
  0x31, 0x32, 0x30, 0x33, 0x30, 0x31, 0x34, 0x36, 0x31, 0x37, 0x5a, 0x30, 0x26, 0x31, 0x24, 0x30,
  0x22, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x1b, 0x43, 0x6c, 0x6f, 0x76, 0x65, 0x72, 0x20, 0x50,
  0x6c, 0x61, 0x74, 0x66, 0x6f, 0x72, 0x6d, 0x20, 0x43, 0x65, 0x72, 0x74, 0x69, 0x66, 0x69, 0x63,
  0x61, 0x74, 0x65, 0x30, 0x82, 0x01, 0x22, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7,
  0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x82, 0x01, 0x0f, 0x00, 0x30, 0x82, 0x01, 0x0a, 0x02,
  0x82, 0x01, 0x01, 0x00, 0xcb, 0xb1, 0x82, 0x68, 0xc1, 0x71, 0xab, 0xf5, 0x35, 0x03, 0xaa, 0x0e,
  0xf7, 0x2e, 0x6d, 0x4b, 0x63, 0x98, 0x8f, 0xd6, 0xae, 0xb0, 0xeb, 0x95, 0x36, 0x70, 0xf9, 0x9d,
  0xe4, 0xf5, 0x6b, 0x0d, 0xc0, 0x81, 0x36, 0x1d, 0xcf, 0x01, 0x45, 0x4c, 0xd6, 0xf0, 0xc5, 0x4e,
  0x7b, 0x4e, 0x49, 0x1a, 0x84, 0x64, 0xc8, 0x7c, 0x87, 0xb8, 0xf3, 0xb2, 0x0a, 0xa1, 0x1e, 0xd6,
  0xb7, 0x42, 0x2d, 0xc4, 0x79, 0x58, 0x19, 0x78, 0x9c, 0x7e, 0x76, 0x8d, 0xfa, 0xb9, 0x6a, 0xdc,
  0x5b, 0xf8, 0x01, 0xa9, 0x53, 0x71, 0xc7, 0x0a, 0x9d, 0x73, 0x2c, 0x4c, 0xdf, 0xe5, 0x88, 0xb4,
  0x10, 0x55, 0xe7, 0xea, 0xce, 0xca, 0xc8, 0xdd, 0x8a, 0xdb, 0x93, 0x43, 0x14, 0x59, 0xb4, 0x04,
  0xab, 0xc7, 0xdc, 0xc7, 0x77, 0xc4, 0x85, 0xf6, 0x02, 0x59, 0x22, 0x91, 0xd0, 0x32, 0x78, 0x47,
  0xf2, 0x2d, 0x0d, 0xd4, 0x81, 0x95, 0x8f, 0x8d, 0xd0, 0x8d, 0x35, 0x1f, 0x7c, 0x2f, 0x44, 0x13,
  0x2b, 0x65, 0x3e, 0x78, 0xba, 0x02, 0x35, 0x2b, 0x56, 0x82, 0x3f, 0x44, 0x56, 0x6b, 0x78, 0x8a,
  0x6e, 0xd7, 0xa2, 0xd2, 0xfd, 0xcc, 0x85, 0x43, 0x8d, 0x73, 0xd8, 0xb4, 0x2f, 0x60, 0xa9, 0x04,
  0xc2, 0xa4, 0xc1, 0x8c, 0xfa, 0x70, 0x0f, 0xe6, 0xdb, 0xcf, 0xbe, 0x95, 0xd2, 0xce, 0xa6, 0x3a,
  0x62, 0xed, 0x6d, 0xc8, 0xa4, 0xa1, 0x8a, 0xe2, 0xb9, 0x01, 0xfd, 0x39, 0x81, 0xa7, 0x2d, 0x6c,
  0xa0, 0x8d, 0xc2, 0x49, 0xdb, 0x3a, 0xc7, 0x46, 0x9b, 0x46, 0x08, 0x17, 0xd4, 0xbb, 0x78, 0xb6,
  0x36, 0x5c, 0xc8, 0x79, 0x9f, 0x62, 0xe7, 0xe2, 0x46, 0x07, 0x32, 0xbc, 0x18, 0xad, 0xf0, 0xb5,
  0x24, 0x34, 0x49, 0xfd, 0xe0, 0x85, 0xcb, 0x22, 0xaa, 0xbc, 0x3a, 0x01, 0x36, 0x1a, 0x1a, 0x07,
  0x05, 0x90, 0xb1, 0xf5, 0x02, 0x03, 0x01, 0x00, 0x01, 0xa3, 0x50, 0x30, 0x4e, 0x30, 0x1d, 0x06,
  0x03, 0x55, 0x1d, 0x0e, 0x04, 0x16, 0x04, 0x14, 0xb4, 0xc6, 0xdc, 0xeb, 0x96, 0x35, 0xe5, 0xe4,
  0xde, 0xf9, 0x53, 0x75, 0x62, 0xf7, 0x7c, 0x79, 0x1a, 0x45, 0x91, 0x3e, 0x30, 0x1f, 0x06, 0x03,
  0x55, 0x1d, 0x23, 0x04, 0x18, 0x30, 0x16, 0x80, 0x14, 0xb4, 0xc6, 0xdc, 0xeb, 0x96, 0x35, 0xe5,
  0xe4, 0xde, 0xf9, 0x53, 0x75, 0x62, 0xf7, 0x7c, 0x79, 0x1a, 0x45, 0x91, 0x3e, 0x30, 0x0c, 0x06,
  0x03, 0x55, 0x1d, 0x13, 0x04, 0x05, 0x30, 0x03, 0x01, 0x01, 0xff, 0x30, 0x0d, 0x06, 0x09, 0x2a,
  0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05, 0x00, 0x03, 0x82, 0x01, 0x01, 0x00, 0x05,
  0xda, 0x57, 0xea, 0x10, 0xfd, 0xe6, 0x19, 0x6c, 0x37, 0xde, 0xac, 0x5a, 0xe0, 0x6a, 0x87, 0x17,
  0x94, 0x80, 0x5e, 0xa6, 0xe8, 0xb8, 0x8c, 0xe6, 0x43, 0xcc, 0x8b, 0x1d, 0x96, 0x87, 0xf2, 0xd9,
  0xfd, 0xc7, 0xb3, 0x1c, 0x67, 0x7e, 0xf0, 0x60, 0x5e, 0x2b, 0xdd, 0x60, 0xc7, 0x18, 0xcc, 0x34,
  0x8e, 0x0b, 0x4e, 0xda, 0x80, 0x30, 0x4b, 0xc9, 0x66, 0xeb, 0x3d, 0x95, 0xf6, 0xac, 0xd8, 0x76,
  0x7e, 0x40, 0x30, 0x6f, 0xb0, 0xb3, 0xf6, 0x7d, 0x13, 0x99, 0x69, 0xdc, 0x6a, 0xd4, 0xd1, 0xc0,
  0x71, 0x69, 0xf3, 0x14, 0x9c, 0x9e, 0xe1, 0xb9, 0x8e, 0x1e, 0xcb, 0x6b, 0x4e, 0xc9, 0xc0, 0x4e,
  0x22, 0x4a, 0xe7, 0xd2, 0x37, 0x33, 0xd0, 0x15, 0xb9, 0x24, 0x04, 0xe8, 0xba, 0x6c, 0x25, 0x29,
  0x67, 0xd1, 0xfb, 0x16, 0x11, 0xca, 0x68, 0xc4, 0x7c, 0x3f, 0x42, 0x6f, 0x7c, 0x59, 0x03, 0xf0,
  0x59, 0x59, 0xd4, 0x72, 0x61, 0x8c, 0xb7, 0xc4, 0xe8, 0xf0, 0x4b, 0x98, 0x45, 0x95, 0x09, 0x77,
  0x17, 0x94, 0x9f, 0xcc, 0xe4, 0xaa, 0x9d, 0xf0, 0x8f, 0xd8, 0x1c, 0x20, 0x2a, 0x92, 0xe0, 0x74,
  0x0d, 0x05, 0xf0, 0xe7, 0x4f, 0x25, 0xd4, 0x72, 0xa1, 0xa4, 0xf1, 0x9a, 0x37, 0x84, 0x1f, 0x75,
  0xb0, 0x8f, 0xc9, 0x71, 0x3e, 0x25, 0x28, 0x96, 0xf6, 0xd8, 0x40, 0xa3, 0x4e, 0xd7, 0xaa, 0x62,
  0xe5, 0x87, 0xac, 0x06, 0x3f, 0x3c, 0x04, 0xe6, 0x64, 0x74, 0x37, 0x05, 0xe6, 0x65, 0xaa, 0xc0,
  0x8f, 0x6c, 0xb7, 0x2f, 0x93, 0xf9, 0x95, 0x6a, 0x9d, 0x3c, 0xc7, 0xbb, 0x20, 0x86, 0x3e, 0x05,
  0x04, 0x2a, 0xc3, 0xe1, 0xcb, 0x5f, 0x32, 0xd7, 0xa3, 0x9f, 0xb3, 0x1d, 0x84, 0xeb, 0x30, 0xeb,
  0x8a, 0x10, 0xf0, 0xc7, 0x38, 0x2c, 0x18, 0x74, 0x2f, 0x76, 0x0a, 0x01, 0xda, 0xa5, 0x4f,
};

// Secure boot null key
STATIC CONST UINT8 gSecureBootPlatformNullSignedKey[] = {
  0xde, 0x07, 0x0b, 0x05, 0x01, 0x37, 0x2c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xc0, 0x05, 0x00, 0x00, 0x00, 0x02, 0xf1, 0x0e, 0x9d, 0xd2, 0xaf, 0x4a, 0xdf, 0x68, 0xee, 0x49,
  0x8a, 0xa9, 0x34, 0x7d, 0x37, 0x56, 0x65, 0xa7, 0x30, 0x82, 0x05, 0xa4, 0x06, 0x09, 0x2a, 0x86,
  0x48, 0x86, 0xf7, 0x0d, 0x01, 0x07, 0x02, 0xa0, 0x82, 0x05, 0x95, 0x30, 0x82, 0x05, 0x91, 0x02,
  0x01, 0x01, 0x31, 0x0f, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02,
  0x01, 0x05, 0x00, 0x30, 0x0b, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x07, 0x01,
  0xa0, 0x82, 0x03, 0x23, 0x30, 0x82, 0x03, 0x1f, 0x30, 0x82, 0x02, 0x07, 0xa0, 0x03, 0x02, 0x01,
  0x02, 0x02, 0x09, 0x00, 0xf0, 0x33, 0x6a, 0xed, 0x40, 0x72, 0xc5, 0x15, 0x30, 0x0d, 0x06, 0x09,
  0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05, 0x00, 0x30, 0x26, 0x31, 0x24, 0x30,
  0x22, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x1b, 0x43, 0x6c, 0x6f, 0x76, 0x65, 0x72, 0x20, 0x50,
  0x6c, 0x61, 0x74, 0x66, 0x6f, 0x72, 0x6d, 0x20, 0x43, 0x65, 0x72, 0x74, 0x69, 0x66, 0x69, 0x63,
  0x61, 0x74, 0x65, 0x30, 0x1e, 0x17, 0x0d, 0x31, 0x33, 0x31, 0x32, 0x30, 0x35, 0x30, 0x31, 0x34,
  0x36, 0x31, 0x37, 0x5a, 0x17, 0x0d, 0x32, 0x33, 0x31, 0x32, 0x30, 0x33, 0x30, 0x31, 0x34, 0x36,
  0x31, 0x37, 0x5a, 0x30, 0x26, 0x31, 0x24, 0x30, 0x22, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x1b,
  0x43, 0x6c, 0x6f, 0x76, 0x65, 0x72, 0x20, 0x50, 0x6c, 0x61, 0x74, 0x66, 0x6f, 0x72, 0x6d, 0x20,
  0x43, 0x65, 0x72, 0x74, 0x69, 0x66, 0x69, 0x63, 0x61, 0x74, 0x65, 0x30, 0x82, 0x01, 0x22, 0x30,
  0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x82,
  0x01, 0x0f, 0x00, 0x30, 0x82, 0x01, 0x0a, 0x02, 0x82, 0x01, 0x01, 0x00, 0xcb, 0xb1, 0x82, 0x68,
  0xc1, 0x71, 0xab, 0xf5, 0x35, 0x03, 0xaa, 0x0e, 0xf7, 0x2e, 0x6d, 0x4b, 0x63, 0x98, 0x8f, 0xd6,
  0xae, 0xb0, 0xeb, 0x95, 0x36, 0x70, 0xf9, 0x9d, 0xe4, 0xf5, 0x6b, 0x0d, 0xc0, 0x81, 0x36, 0x1d,
  0xcf, 0x01, 0x45, 0x4c, 0xd6, 0xf0, 0xc5, 0x4e, 0x7b, 0x4e, 0x49, 0x1a, 0x84, 0x64, 0xc8, 0x7c,
  0x87, 0xb8, 0xf3, 0xb2, 0x0a, 0xa1, 0x1e, 0xd6, 0xb7, 0x42, 0x2d, 0xc4, 0x79, 0x58, 0x19, 0x78,
  0x9c, 0x7e, 0x76, 0x8d, 0xfa, 0xb9, 0x6a, 0xdc, 0x5b, 0xf8, 0x01, 0xa9, 0x53, 0x71, 0xc7, 0x0a,
  0x9d, 0x73, 0x2c, 0x4c, 0xdf, 0xe5, 0x88, 0xb4, 0x10, 0x55, 0xe7, 0xea, 0xce, 0xca, 0xc8, 0xdd,
  0x8a, 0xdb, 0x93, 0x43, 0x14, 0x59, 0xb4, 0x04, 0xab, 0xc7, 0xdc, 0xc7, 0x77, 0xc4, 0x85, 0xf6,
  0x02, 0x59, 0x22, 0x91, 0xd0, 0x32, 0x78, 0x47, 0xf2, 0x2d, 0x0d, 0xd4, 0x81, 0x95, 0x8f, 0x8d,
  0xd0, 0x8d, 0x35, 0x1f, 0x7c, 0x2f, 0x44, 0x13, 0x2b, 0x65, 0x3e, 0x78, 0xba, 0x02, 0x35, 0x2b,
  0x56, 0x82, 0x3f, 0x44, 0x56, 0x6b, 0x78, 0x8a, 0x6e, 0xd7, 0xa2, 0xd2, 0xfd, 0xcc, 0x85, 0x43,
  0x8d, 0x73, 0xd8, 0xb4, 0x2f, 0x60, 0xa9, 0x04, 0xc2, 0xa4, 0xc1, 0x8c, 0xfa, 0x70, 0x0f, 0xe6,
  0xdb, 0xcf, 0xbe, 0x95, 0xd2, 0xce, 0xa6, 0x3a, 0x62, 0xed, 0x6d, 0xc8, 0xa4, 0xa1, 0x8a, 0xe2,
  0xb9, 0x01, 0xfd, 0x39, 0x81, 0xa7, 0x2d, 0x6c, 0xa0, 0x8d, 0xc2, 0x49, 0xdb, 0x3a, 0xc7, 0x46,
  0x9b, 0x46, 0x08, 0x17, 0xd4, 0xbb, 0x78, 0xb6, 0x36, 0x5c, 0xc8, 0x79, 0x9f, 0x62, 0xe7, 0xe2,
  0x46, 0x07, 0x32, 0xbc, 0x18, 0xad, 0xf0, 0xb5, 0x24, 0x34, 0x49, 0xfd, 0xe0, 0x85, 0xcb, 0x22,
  0xaa, 0xbc, 0x3a, 0x01, 0x36, 0x1a, 0x1a, 0x07, 0x05, 0x90, 0xb1, 0xf5, 0x02, 0x03, 0x01, 0x00,
  0x01, 0xa3, 0x50, 0x30, 0x4e, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x1d, 0x0e, 0x04, 0x16, 0x04, 0x14,
  0xb4, 0xc6, 0xdc, 0xeb, 0x96, 0x35, 0xe5, 0xe4, 0xde, 0xf9, 0x53, 0x75, 0x62, 0xf7, 0x7c, 0x79,
  0x1a, 0x45, 0x91, 0x3e, 0x30, 0x1f, 0x06, 0x03, 0x55, 0x1d, 0x23, 0x04, 0x18, 0x30, 0x16, 0x80,
  0x14, 0xb4, 0xc6, 0xdc, 0xeb, 0x96, 0x35, 0xe5, 0xe4, 0xde, 0xf9, 0x53, 0x75, 0x62, 0xf7, 0x7c,
  0x79, 0x1a, 0x45, 0x91, 0x3e, 0x30, 0x0c, 0x06, 0x03, 0x55, 0x1d, 0x13, 0x04, 0x05, 0x30, 0x03,
  0x01, 0x01, 0xff, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b,
  0x05, 0x00, 0x03, 0x82, 0x01, 0x01, 0x00, 0x05, 0xda, 0x57, 0xea, 0x10, 0xfd, 0xe6, 0x19, 0x6c,
  0x37, 0xde, 0xac, 0x5a, 0xe0, 0x6a, 0x87, 0x17, 0x94, 0x80, 0x5e, 0xa6, 0xe8, 0xb8, 0x8c, 0xe6,
  0x43, 0xcc, 0x8b, 0x1d, 0x96, 0x87, 0xf2, 0xd9, 0xfd, 0xc7, 0xb3, 0x1c, 0x67, 0x7e, 0xf0, 0x60,
  0x5e, 0x2b, 0xdd, 0x60, 0xc7, 0x18, 0xcc, 0x34, 0x8e, 0x0b, 0x4e, 0xda, 0x80, 0x30, 0x4b, 0xc9,
  0x66, 0xeb, 0x3d, 0x95, 0xf6, 0xac, 0xd8, 0x76, 0x7e, 0x40, 0x30, 0x6f, 0xb0, 0xb3, 0xf6, 0x7d,
  0x13, 0x99, 0x69, 0xdc, 0x6a, 0xd4, 0xd1, 0xc0, 0x71, 0x69, 0xf3, 0x14, 0x9c, 0x9e, 0xe1, 0xb9,
  0x8e, 0x1e, 0xcb, 0x6b, 0x4e, 0xc9, 0xc0, 0x4e, 0x22, 0x4a, 0xe7, 0xd2, 0x37, 0x33, 0xd0, 0x15,
  0xb9, 0x24, 0x04, 0xe8, 0xba, 0x6c, 0x25, 0x29, 0x67, 0xd1, 0xfb, 0x16, 0x11, 0xca, 0x68, 0xc4,
  0x7c, 0x3f, 0x42, 0x6f, 0x7c, 0x59, 0x03, 0xf0, 0x59, 0x59, 0xd4, 0x72, 0x61, 0x8c, 0xb7, 0xc4,
  0xe8, 0xf0, 0x4b, 0x98, 0x45, 0x95, 0x09, 0x77, 0x17, 0x94, 0x9f, 0xcc, 0xe4, 0xaa, 0x9d, 0xf0,
  0x8f, 0xd8, 0x1c, 0x20, 0x2a, 0x92, 0xe0, 0x74, 0x0d, 0x05, 0xf0, 0xe7, 0x4f, 0x25, 0xd4, 0x72,
  0xa1, 0xa4, 0xf1, 0x9a, 0x37, 0x84, 0x1f, 0x75, 0xb0, 0x8f, 0xc9, 0x71, 0x3e, 0x25, 0x28, 0x96,
  0xf6, 0xd8, 0x40, 0xa3, 0x4e, 0xd7, 0xaa, 0x62, 0xe5, 0x87, 0xac, 0x06, 0x3f, 0x3c, 0x04, 0xe6,
  0x64, 0x74, 0x37, 0x05, 0xe6, 0x65, 0xaa, 0xc0, 0x8f, 0x6c, 0xb7, 0x2f, 0x93, 0xf9, 0x95, 0x6a,
  0x9d, 0x3c, 0xc7, 0xbb, 0x20, 0x86, 0x3e, 0x05, 0x04, 0x2a, 0xc3, 0xe1, 0xcb, 0x5f, 0x32, 0xd7,
  0xa3, 0x9f, 0xb3, 0x1d, 0x84, 0xeb, 0x30, 0xeb, 0x8a, 0x10, 0xf0, 0xc7, 0x38, 0x2c, 0x18, 0x74,
  0x2f, 0x76, 0x0a, 0x01, 0xda, 0xa5, 0x4f, 0x31, 0x82, 0x02, 0x45, 0x30, 0x82, 0x02, 0x41, 0x02,
  0x01, 0x01, 0x30, 0x33, 0x30, 0x26, 0x31, 0x24, 0x30, 0x22, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c,
  0x1b, 0x43, 0x6c, 0x6f, 0x76, 0x65, 0x72, 0x20, 0x50, 0x6c, 0x61, 0x74, 0x66, 0x6f, 0x72, 0x6d,
  0x20, 0x43, 0x65, 0x72, 0x74, 0x69, 0x66, 0x69, 0x63, 0x61, 0x74, 0x65, 0x02, 0x09, 0x00, 0xf0,
  0x33, 0x6a, 0xed, 0x40, 0x72, 0xc5, 0x15, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65,
  0x03, 0x04, 0x02, 0x01, 0x05, 0x00, 0xa0, 0x81, 0xe4, 0x30, 0x18, 0x06, 0x09, 0x2a, 0x86, 0x48,
  0x86, 0xf7, 0x0d, 0x01, 0x09, 0x03, 0x31, 0x0b, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
  0x01, 0x07, 0x01, 0x30, 0x1c, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x09, 0x05,
  0x31, 0x0f, 0x17, 0x0d, 0x31, 0x33, 0x31, 0x32, 0x30, 0x35, 0x30, 0x31, 0x35, 0x35, 0x34, 0x34,
  0x5a, 0x30, 0x2f, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x09, 0x04, 0x31, 0x22,
  0x04, 0x20, 0xc7, 0x14, 0x77, 0x0e, 0xc4, 0xeb, 0x54, 0x8c, 0xf5, 0x23, 0x81, 0x25, 0x23, 0xd5,
  0x31, 0x11, 0x3d, 0xdf, 0xe9, 0xaf, 0x8d, 0x76, 0xa1, 0xc0, 0xa2, 0xc6, 0x49, 0x82, 0x6a, 0xc9,
  0x1e, 0xae, 0x30, 0x79, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x09, 0x0f, 0x31,
  0x6c, 0x30, 0x6a, 0x30, 0x0b, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x01, 0x2a,
  0x30, 0x0b, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x01, 0x16, 0x30, 0x0b, 0x06,
  0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x01, 0x02, 0x30, 0x0a, 0x06, 0x08, 0x2a, 0x86,
  0x48, 0x86, 0xf7, 0x0d, 0x03, 0x07, 0x30, 0x0e, 0x06, 0x08, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
  0x03, 0x02, 0x02, 0x02, 0x00, 0x80, 0x30, 0x0d, 0x06, 0x08, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
  0x03, 0x02, 0x02, 0x01, 0x40, 0x30, 0x07, 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 0x07, 0x30, 0x0d,
  0x06, 0x08, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x03, 0x02, 0x02, 0x01, 0x28, 0x30, 0x0d, 0x06,
  0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x04, 0x82, 0x01, 0x00,
  0x2d, 0x05, 0x15, 0x6b, 0x79, 0xdf, 0x3e, 0x09, 0x8c, 0xf5, 0x4e, 0x47, 0xe1, 0x87, 0x92, 0xa1,
  0x60, 0xb7, 0x59, 0x74, 0xaa, 0x49, 0x72, 0x08, 0x09, 0x61, 0x40, 0x35, 0x25, 0x8f, 0xba, 0x68,
  0x72, 0x95, 0xc5, 0x70, 0x2c, 0x50, 0x5b, 0xb5, 0x8d, 0x6d, 0x1d, 0x71, 0x4a, 0x18, 0x85, 0x09,
  0x90, 0x05, 0x3c, 0x75, 0xbf, 0x71, 0xc1, 0x0a, 0xb5, 0x9b, 0x1a, 0xd6, 0xe4, 0xd9, 0xca, 0x60,
  0x7e, 0x77, 0x8d, 0x1e, 0x4e, 0xc8, 0x83, 0xe0, 0x65, 0x6b, 0x8c, 0x24, 0x74, 0xe5, 0x4c, 0xc1,
  0xc8, 0x99, 0x99, 0xa9, 0xed, 0x44, 0xae, 0x11, 0x99, 0xed, 0xe8, 0x91, 0x84, 0x02, 0x4b, 0x2a,
  0x00, 0x5f, 0xe4, 0xb0, 0x42, 0xb1, 0xb3, 0x0d, 0x46, 0xd9, 0xc8, 0x5e, 0x40, 0xf6, 0xea, 0x8c,
  0xb0, 0xfe, 0x68, 0x7e, 0x24, 0x23, 0x76, 0xe1, 0xf0, 0xd1, 0xf7, 0x03, 0xb0, 0xd5, 0xf5, 0x9c,
  0xc6, 0x19, 0x51, 0x46, 0x0e, 0x68, 0x35, 0xfa, 0xbf, 0x34, 0x04, 0xd8, 0xc7, 0x9c, 0xd1, 0xdf,
  0x6b, 0x1b, 0x00, 0xd5, 0x4a, 0x7a, 0x8f, 0x69, 0x65, 0x8d, 0x56, 0xed, 0xed, 0x95, 0xd4, 0xe0,
  0x7f, 0x4a, 0x8c, 0x82, 0x51, 0x22, 0x11, 0x3f, 0x9e, 0x77, 0xb6, 0x27, 0x8e, 0xff, 0x23, 0x77,
  0xcd, 0x72, 0xc6, 0xa9, 0xf5, 0x66, 0x0a, 0x24, 0x5d, 0x80, 0xb5, 0x3a, 0x12, 0x29, 0x9e, 0x5f,
  0xed, 0xc1, 0xb4, 0xd5, 0x37, 0x9a, 0xa4, 0x3e, 0xe0, 0x4f, 0x7f, 0x85, 0xca, 0x6f, 0x9f, 0xd2,
  0x29, 0xfb, 0xdb, 0x0d, 0x2d, 0x66, 0x85, 0x91, 0xc7, 0x14, 0x6a, 0x8e, 0x1c, 0xac, 0x4a, 0x79,
  0xa2, 0x75, 0x81, 0xd2, 0x46, 0x9b, 0x17, 0x78, 0x8e, 0x39, 0xf8, 0xf9, 0xe7, 0xfb, 0x79, 0x2e,
  0x3d, 0x12, 0x68, 0xf6, 0x84, 0xff, 0x47, 0x01, 0x9b, 0x90, 0x00, 0x17, 0x01, 0x1a, 0xa8, 0xb9,
};

// Secure boot exchange key
STATIC CONST UINT8 gSecureBootExchangeKey[] = {
  0x2d, 0x2d, 0x2d, 0x2d, 0x2d, 0x42, 0x45, 0x47, 0x49, 0x4e, 0x20, 0x43, 0x45, 0x52, 0x54, 0x49,
  0x46, 0x49, 0x43, 0x41, 0x54, 0x45, 0x2d, 0x2d, 0x2d, 0x2d, 0x2d, 0x0a, 0x4d, 0x49, 0x49, 0x44,
  0x48, 0x7a, 0x43, 0x43, 0x41, 0x67, 0x65, 0x67, 0x41, 0x77, 0x49, 0x42, 0x41, 0x67, 0x49, 0x4a,
  0x41, 0x49, 0x73, 0x55, 0x75, 0x63, 0x54, 0x53, 0x45, 0x4c, 0x64, 0x75, 0x4d, 0x41, 0x30, 0x47,
  0x43, 0x53, 0x71, 0x47, 0x53, 0x49, 0x62, 0x33, 0x44, 0x51, 0x45, 0x42, 0x43, 0x77, 0x55, 0x41,
  0x4d, 0x43, 0x59, 0x78, 0x4a, 0x44, 0x41, 0x69, 0x42, 0x67, 0x4e, 0x56, 0x0a, 0x42, 0x41, 0x4d,
  0x4d, 0x47, 0x30, 0x4e, 0x73, 0x62, 0x33, 0x5a, 0x6c, 0x63, 0x69, 0x42, 0x46, 0x65, 0x47, 0x4e,
  0x6f, 0x59, 0x57, 0x35, 0x6e, 0x5a, 0x53, 0x42, 0x44, 0x5a, 0x58, 0x4a, 0x30, 0x61, 0x57, 0x5a,
  0x70, 0x59, 0x32, 0x46, 0x30, 0x5a, 0x54, 0x41, 0x65, 0x46, 0x77, 0x30, 0x78, 0x4d, 0x7a, 0x45,
  0x79, 0x4d, 0x44, 0x55, 0x77, 0x4d, 0x54, 0x51, 0x32, 0x4e, 0x44, 0x52, 0x61, 0x0a, 0x46, 0x77,
  0x30, 0x79, 0x4d, 0x7a, 0x45, 0x79, 0x4d, 0x44, 0x4d, 0x77, 0x4d, 0x54, 0x51, 0x32, 0x4e, 0x44,
  0x52, 0x61, 0x4d, 0x43, 0x59, 0x78, 0x4a, 0x44, 0x41, 0x69, 0x42, 0x67, 0x4e, 0x56, 0x42, 0x41,
  0x4d, 0x4d, 0x47, 0x30, 0x4e, 0x73, 0x62, 0x33, 0x5a, 0x6c, 0x63, 0x69, 0x42, 0x46, 0x65, 0x47,
  0x4e, 0x6f, 0x59, 0x57, 0x35, 0x6e, 0x5a, 0x53, 0x42, 0x44, 0x5a, 0x58, 0x4a, 0x30, 0x0a, 0x61,
  0x57, 0x5a, 0x70, 0x59, 0x32, 0x46, 0x30, 0x5a, 0x54, 0x43, 0x43, 0x41, 0x53, 0x49, 0x77, 0x44,
  0x51, 0x59, 0x4a, 0x4b, 0x6f, 0x5a, 0x49, 0x68, 0x76, 0x63, 0x4e, 0x41, 0x51, 0x45, 0x42, 0x42,
  0x51, 0x41, 0x44, 0x67, 0x67, 0x45, 0x50, 0x41, 0x44, 0x43, 0x43, 0x41, 0x51, 0x6f, 0x43, 0x67,
  0x67, 0x45, 0x42, 0x41, 0x4e, 0x49, 0x6d, 0x4a, 0x61, 0x4b, 0x44, 0x67, 0x76, 0x66, 0x58, 0x0a,
  0x37, 0x68, 0x62, 0x46, 0x2f, 0x42, 0x38, 0x59, 0x6f, 0x75, 0x6f, 0x36, 0x76, 0x6a, 0x31, 0x64,
  0x54, 0x49, 0x57, 0x34, 0x42, 0x4c, 0x77, 0x49, 0x78, 0x4f, 0x55, 0x4a, 0x4e, 0x78, 0x78, 0x76,
  0x31, 0x4d, 0x4e, 0x79, 0x2f, 0x77, 0x48, 0x53, 0x2f, 0x61, 0x44, 0x2f, 0x66, 0x55, 0x55, 0x70,
  0x6a, 0x4b, 0x79, 0x62, 0x41, 0x38, 0x61, 0x61, 0x7a, 0x62, 0x36, 0x61, 0x6d, 0x75, 0x64, 0x76,
  0x0a, 0x51, 0x38, 0x66, 0x47, 0x6c, 0x4f, 0x36, 0x51, 0x4c, 0x62, 0x37, 0x53, 0x77, 0x62, 0x57,
  0x6a, 0x54, 0x5a, 0x2b, 0x4e, 0x59, 0x51, 0x56, 0x30, 0x58, 0x33, 0x70, 0x56, 0x33, 0x64, 0x31,
  0x52, 0x49, 0x69, 0x72, 0x4b, 0x49, 0x36, 0x50, 0x44, 0x38, 0x2f, 0x48, 0x51, 0x63, 0x71, 0x4d,
  0x78, 0x6b, 0x79, 0x66, 0x36, 0x47, 0x4c, 0x70, 0x61, 0x6c, 0x6f, 0x33, 0x30, 0x34, 0x55, 0x31,
  0x2f, 0x0a, 0x33, 0x33, 0x65, 0x36, 0x6f, 0x32, 0x37, 0x4a, 0x79, 0x68, 0x4a, 0x64, 0x78, 0x44,
  0x4e, 0x68, 0x4d, 0x56, 0x36, 0x77, 0x37, 0x4b, 0x69, 0x42, 0x7a, 0x68, 0x37, 0x6f, 0x6b, 0x32,
  0x55, 0x47, 0x57, 0x45, 0x65, 0x42, 0x72, 0x2b, 0x6e, 0x74, 0x5a, 0x7a, 0x50, 0x35, 0x41, 0x67,
  0x31, 0x6e, 0x4a, 0x4e, 0x4e, 0x6e, 0x79, 0x78, 0x35, 0x54, 0x55, 0x52, 0x71, 0x66, 0x33, 0x74,
  0x55, 0x37, 0x0a, 0x32, 0x52, 0x4a, 0x67, 0x71, 0x63, 0x65, 0x38, 0x44, 0x30, 0x58, 0x53, 0x73,
  0x58, 0x78, 0x52, 0x46, 0x59, 0x4b, 0x7a, 0x48, 0x30, 0x5a, 0x36, 0x7a, 0x75, 0x74, 0x35, 0x2f,
  0x47, 0x31, 0x34, 0x30, 0x79, 0x44, 0x54, 0x6e, 0x48, 0x41, 0x77, 0x62, 0x35, 0x45, 0x58, 0x30,
  0x35, 0x30, 0x64, 0x34, 0x74, 0x4c, 0x65, 0x54, 0x43, 0x6d, 0x42, 0x37, 0x30, 0x31, 0x33, 0x61,
  0x35, 0x52, 0x55, 0x0a, 0x57, 0x6a, 0x41, 0x4a, 0x31, 0x43, 0x73, 0x6a, 0x6b, 0x51, 0x65, 0x42,
  0x35, 0x4c, 0x72, 0x39, 0x75, 0x4d, 0x45, 0x63, 0x4f, 0x52, 0x48, 0x49, 0x36, 0x6d, 0x75, 0x79,
  0x44, 0x57, 0x47, 0x4a, 0x30, 0x44, 0x4d, 0x6e, 0x73, 0x6e, 0x32, 0x48, 0x69, 0x77, 0x4c, 0x6f,
  0x35, 0x5a, 0x66, 0x30, 0x62, 0x77, 0x69, 0x42, 0x42, 0x34, 0x36, 0x70, 0x74, 0x63, 0x76, 0x44,
  0x4a, 0x45, 0x69, 0x58, 0x0a, 0x42, 0x64, 0x31, 0x64, 0x65, 0x59, 0x36, 0x75, 0x64, 0x59, 0x6b,
  0x43, 0x41, 0x77, 0x45, 0x41, 0x41, 0x61, 0x4e, 0x51, 0x4d, 0x45, 0x34, 0x77, 0x48, 0x51, 0x59,
  0x44, 0x56, 0x52, 0x30, 0x4f, 0x42, 0x42, 0x59, 0x45, 0x46, 0x44, 0x46, 0x70, 0x52, 0x70, 0x65,
  0x69, 0x2b, 0x36, 0x59, 0x35, 0x7a, 0x4f, 0x76, 0x57, 0x51, 0x31, 0x57, 0x6e, 0x71, 0x44, 0x33,
  0x6b, 0x78, 0x65, 0x37, 0x74, 0x0a, 0x4d, 0x42, 0x38, 0x47, 0x41, 0x31, 0x55, 0x64, 0x49, 0x77,
  0x51, 0x59, 0x4d, 0x42, 0x61, 0x41, 0x46, 0x44, 0x46, 0x70, 0x52, 0x70, 0x65, 0x69, 0x2b, 0x36,
  0x59, 0x35, 0x7a, 0x4f, 0x76, 0x57, 0x51, 0x31, 0x57, 0x6e, 0x71, 0x44, 0x33, 0x6b, 0x78, 0x65,
  0x37, 0x74, 0x4d, 0x41, 0x77, 0x47, 0x41, 0x31, 0x55, 0x64, 0x45, 0x77, 0x51, 0x46, 0x4d, 0x41,
  0x4d, 0x42, 0x41, 0x66, 0x38, 0x77, 0x0a, 0x44, 0x51, 0x59, 0x4a, 0x4b, 0x6f, 0x5a, 0x49, 0x68,
  0x76, 0x63, 0x4e, 0x41, 0x51, 0x45, 0x4c, 0x42, 0x51, 0x41, 0x44, 0x67, 0x67, 0x45, 0x42, 0x41,
  0x4b, 0x4e, 0x58, 0x42, 0x35, 0x48, 0x6e, 0x75, 0x41, 0x49, 0x45, 0x70, 0x35, 0x56, 0x31, 0x31,
  0x30, 0x61, 0x58, 0x78, 0x62, 0x33, 0x37, 0x50, 0x50, 0x33, 0x38, 0x53, 0x6a, 0x65, 0x6f, 0x77,
  0x49, 0x4b, 0x35, 0x51, 0x66, 0x6f, 0x76, 0x0a, 0x70, 0x58, 0x31, 0x59, 0x50, 0x6c, 0x6a, 0x32,
  0x59, 0x71, 0x37, 0x58, 0x6c, 0x69, 0x79, 0x67, 0x76, 0x58, 0x50, 0x68, 0x64, 0x75, 0x46, 0x6f,
  0x41, 0x4b, 0x77, 0x6a, 0x2f, 0x76, 0x59, 0x4c, 0x70, 0x66, 0x6a, 0x5a, 0x4a, 0x66, 0x42, 0x48,
  0x35, 0x35, 0x6e, 0x58, 0x37, 0x57, 0x72, 0x5a, 0x6b, 0x57, 0x64, 0x65, 0x41, 0x4e, 0x44, 0x47,
  0x34, 0x4a, 0x30, 0x75, 0x36, 0x4a, 0x65, 0x76, 0x0a, 0x35, 0x64, 0x44, 0x4f, 0x34, 0x46, 0x56,
  0x73, 0x74, 0x2f, 0x7a, 0x66, 0x56, 0x4b, 0x4c, 0x52, 0x6f, 0x58, 0x59, 0x57, 0x35, 0x74, 0x6c,
  0x7a, 0x79, 0x35, 0x4e, 0x79, 0x56, 0x47, 0x5a, 0x30, 0x31, 0x73, 0x42, 0x35, 0x47, 0x78, 0x4a,
  0x58, 0x59, 0x74, 0x54, 0x4d, 0x61, 0x58, 0x47, 0x37, 0x30, 0x4e, 0x68, 0x69, 0x46, 0x45, 0x2b,
  0x69, 0x42, 0x31, 0x6d, 0x6e, 0x33, 0x63, 0x6c, 0x35, 0x0a, 0x54, 0x4b, 0x69, 0x39, 0x45, 0x65,
  0x30, 0x39, 0x48, 0x45, 0x54, 0x4e, 0x4e, 0x57, 0x55, 0x71, 0x75, 0x61, 0x76, 0x7a, 0x6f, 0x66,
  0x4b, 0x67, 0x79, 0x2b, 0x7a, 0x74, 0x62, 0x31, 0x68, 0x39, 0x36, 0x72, 0x2b, 0x54, 0x2b, 0x66,
  0x4a, 0x4a, 0x46, 0x64, 0x51, 0x36, 0x5a, 0x50, 0x69, 0x70, 0x79, 0x59, 0x32, 0x2b, 0x67, 0x53,
  0x55, 0x41, 0x38, 0x72, 0x2b, 0x34, 0x57, 0x6e, 0x4d, 0x73, 0x0a, 0x50, 0x30, 0x37, 0x77, 0x43,
  0x74, 0x6d, 0x54, 0x59, 0x34, 0x52, 0x2b, 0x33, 0x64, 0x51, 0x5a, 0x72, 0x2f, 0x31, 0x2f, 0x50,
  0x44, 0x56, 0x58, 0x4d, 0x73, 0x5a, 0x38, 0x39, 0x48, 0x61, 0x44, 0x64, 0x30, 0x57, 0x52, 0x4c,
  0x63, 0x37, 0x55, 0x41, 0x69, 0x45, 0x50, 0x2b, 0x50, 0x34, 0x73, 0x79, 0x6b, 0x59, 0x75, 0x6c,
  0x35, 0x39, 0x62, 0x5a, 0x50, 0x6a, 0x7a, 0x41, 0x49, 0x70, 0x68, 0x0a, 0x46, 0x33, 0x52, 0x43,
  0x51, 0x78, 0x4a, 0x57, 0x32, 0x4b, 0x4b, 0x6f, 0x47, 0x68, 0x6e, 0x39, 0x51, 0x6b, 0x6e, 0x56,
  0x2f, 0x58, 0x73, 0x79, 0x62, 0x2b, 0x52, 0x42, 0x30, 0x35, 0x46, 0x4b, 0x41, 0x65, 0x43, 0x6c,
  0x6e, 0x43, 0x56, 0x64, 0x4a, 0x78, 0x38, 0x57, 0x6c, 0x75, 0x45, 0x3d, 0x0a, 0x2d, 0x2d, 0x2d,
  0x2d, 0x2d, 0x45, 0x4e, 0x44, 0x20, 0x43, 0x45, 0x52, 0x54, 0x49, 0x46, 0x49, 0x43, 0x41, 0x54,
  0x45, 0x2d, 0x2d, 0x2d, 0x2d, 0x2d, 0x0a,
};

// Secure boot private exchange key
STATIC CONST UINT8 gSecureBootExchangePrivateKey[] = {
  0x2d, 0x2d, 0x2d, 0x2d, 0x2d, 0x42, 0x45, 0x47, 0x49, 0x4e, 0x20, 0x50, 0x52, 0x49, 0x56, 0x41,
  0x54, 0x45, 0x20, 0x4b, 0x45, 0x59, 0x2d, 0x2d, 0x2d, 0x2d, 0x2d, 0x0a, 0x4d, 0x49, 0x49, 0x45,
  0x76, 0x51, 0x49, 0x42, 0x41, 0x44, 0x41, 0x4e, 0x42, 0x67, 0x6b, 0x71, 0x68, 0x6b, 0x69, 0x47,
  0x39, 0x77, 0x30, 0x42, 0x41, 0x51, 0x45, 0x46, 0x41, 0x41, 0x53, 0x43, 0x42, 0x4b, 0x63, 0x77,
  0x67, 0x67, 0x53, 0x6a, 0x41, 0x67, 0x45, 0x41, 0x41, 0x6f, 0x49, 0x42, 0x41, 0x51, 0x44, 0x53,
  0x4a, 0x69, 0x57, 0x69, 0x67, 0x34, 0x4c, 0x33, 0x31, 0x2b, 0x34, 0x57, 0x0a, 0x78, 0x66, 0x77,
  0x66, 0x47, 0x4b, 0x4c, 0x71, 0x4f, 0x72, 0x34, 0x39, 0x58, 0x55, 0x79, 0x46, 0x75, 0x41, 0x53,
  0x38, 0x43, 0x4d, 0x54, 0x6c, 0x43, 0x54, 0x63, 0x63, 0x62, 0x39, 0x54, 0x44, 0x63, 0x76, 0x38,
  0x42, 0x30, 0x76, 0x32, 0x67, 0x2f, 0x33, 0x31, 0x46, 0x4b, 0x59, 0x79, 0x73, 0x6d, 0x77, 0x50,
  0x47, 0x6d, 0x73, 0x32, 0x2b, 0x6d, 0x70, 0x72, 0x6e, 0x62, 0x30, 0x50, 0x48, 0x0a, 0x78, 0x70,
  0x54, 0x75, 0x6b, 0x43, 0x32, 0x2b, 0x30, 0x73, 0x47, 0x31, 0x6f, 0x30, 0x32, 0x66, 0x6a, 0x57,
  0x45, 0x46, 0x64, 0x46, 0x39, 0x36, 0x56, 0x64, 0x33, 0x64, 0x55, 0x53, 0x49, 0x71, 0x79, 0x69,
  0x4f, 0x6a, 0x77, 0x2f, 0x50, 0x78, 0x30, 0x48, 0x4b, 0x6a, 0x4d, 0x5a, 0x4d, 0x6e, 0x2b, 0x68,
  0x69, 0x36, 0x57, 0x70, 0x61, 0x4e, 0x39, 0x4f, 0x46, 0x4e, 0x66, 0x39, 0x39, 0x33, 0x0a, 0x75,
  0x71, 0x4e, 0x75, 0x79, 0x63, 0x6f, 0x53, 0x58, 0x63, 0x51, 0x7a, 0x59, 0x54, 0x46, 0x65, 0x73,
  0x4f, 0x79, 0x6f, 0x67, 0x63, 0x34, 0x65, 0x36, 0x4a, 0x4e, 0x6c, 0x42, 0x6c, 0x68, 0x48, 0x67,
  0x61, 0x2f, 0x70, 0x37, 0x57, 0x63, 0x7a, 0x2b, 0x51, 0x49, 0x4e, 0x5a, 0x79, 0x54, 0x54, 0x5a,
  0x38, 0x73, 0x65, 0x55, 0x31, 0x45, 0x61, 0x6e, 0x39, 0x37, 0x56, 0x4f, 0x39, 0x6b, 0x53, 0x0a,
  0x59, 0x4b, 0x6e, 0x48, 0x76, 0x41, 0x39, 0x46, 0x30, 0x72, 0x46, 0x38, 0x55, 0x52, 0x57, 0x43,
  0x73, 0x78, 0x39, 0x47, 0x65, 0x73, 0x37, 0x72, 0x65, 0x66, 0x78, 0x74, 0x65, 0x4e, 0x4d, 0x67,
  0x30, 0x35, 0x78, 0x77, 0x4d, 0x47, 0x2b, 0x52, 0x46, 0x39, 0x4f, 0x64, 0x48, 0x65, 0x4c, 0x53,
  0x33, 0x6b, 0x77, 0x70, 0x67, 0x65, 0x39, 0x4e, 0x64, 0x32, 0x75, 0x55, 0x56, 0x46, 0x6f, 0x77,
  0x0a, 0x43, 0x64, 0x51, 0x72, 0x49, 0x35, 0x45, 0x48, 0x67, 0x65, 0x53, 0x36, 0x2f, 0x62, 0x6a,
  0x42, 0x48, 0x44, 0x6b, 0x52, 0x79, 0x4f, 0x70, 0x72, 0x73, 0x67, 0x31, 0x68, 0x69, 0x64, 0x41,
  0x7a, 0x4a, 0x37, 0x4a, 0x39, 0x68, 0x34, 0x73, 0x43, 0x36, 0x4f, 0x57, 0x58, 0x39, 0x47, 0x38,
  0x49, 0x67, 0x51, 0x65, 0x4f, 0x71, 0x62, 0x58, 0x4c, 0x77, 0x79, 0x52, 0x49, 0x6c, 0x77, 0x58,
  0x64, 0x0a, 0x58, 0x58, 0x6d, 0x4f, 0x72, 0x6e, 0x57, 0x4a, 0x41, 0x67, 0x4d, 0x42, 0x41, 0x41,
  0x45, 0x43, 0x67, 0x67, 0x45, 0x41, 0x58, 0x6d, 0x2f, 0x4d, 0x47, 0x39, 0x4c, 0x7a, 0x71, 0x51,
  0x4c, 0x39, 0x4f, 0x48, 0x69, 0x79, 0x53, 0x78, 0x73, 0x63, 0x49, 0x78, 0x78, 0x51, 0x33, 0x32,
  0x79, 0x73, 0x39, 0x61, 0x35, 0x7a, 0x57, 0x6a, 0x69, 0x71, 0x6d, 0x71, 0x35, 0x67, 0x65, 0x4d,
  0x50, 0x38, 0x0a, 0x6e, 0x66, 0x52, 0x77, 0x69, 0x6c, 0x5a, 0x52, 0x71, 0x65, 0x63, 0x47, 0x57,
  0x57, 0x31, 0x4a, 0x6a, 0x55, 0x75, 0x41, 0x44, 0x2b, 0x45, 0x44, 0x4d, 0x6e, 0x4e, 0x78, 0x78,
  0x79, 0x30, 0x6d, 0x44, 0x59, 0x47, 0x38, 0x37, 0x64, 0x51, 0x57, 0x36, 0x70, 0x6a, 0x31, 0x65,
  0x66, 0x57, 0x76, 0x37, 0x56, 0x4a, 0x77, 0x56, 0x69, 0x34, 0x73, 0x74, 0x65, 0x49, 0x65, 0x65,
  0x4e, 0x56, 0x38, 0x0a, 0x41, 0x44, 0x53, 0x6b, 0x66, 0x58, 0x53, 0x53, 0x6a, 0x49, 0x6c, 0x36,
  0x36, 0x73, 0x46, 0x64, 0x65, 0x75, 0x71, 0x56, 0x38, 0x44, 0x44, 0x44, 0x56, 0x58, 0x6e, 0x46,
  0x51, 0x62, 0x6d, 0x45, 0x4a, 0x39, 0x37, 0x55, 0x6b, 0x50, 0x58, 0x58, 0x66, 0x42, 0x64, 0x74,
  0x45, 0x78, 0x58, 0x4c, 0x52, 0x4e, 0x31, 0x48, 0x44, 0x4a, 0x61, 0x41, 0x6c, 0x54, 0x30, 0x34,
  0x39, 0x33, 0x31, 0x6f, 0x0a, 0x49, 0x6c, 0x52, 0x68, 0x7a, 0x4d, 0x64, 0x79, 0x7a, 0x31, 0x6b,
  0x6e, 0x30, 0x6c, 0x57, 0x76, 0x47, 0x66, 0x6e, 0x31, 0x45, 0x61, 0x30, 0x39, 0x62, 0x49, 0x54,
  0x71, 0x53, 0x7a, 0x43, 0x6d, 0x4f, 0x70, 0x34, 0x66, 0x57, 0x37, 0x30, 0x32, 0x2b, 0x32, 0x77,
  0x33, 0x64, 0x46, 0x37, 0x34, 0x75, 0x70, 0x38, 0x38, 0x4a, 0x64, 0x62, 0x74, 0x46, 0x52, 0x68,
  0x78, 0x62, 0x34, 0x45, 0x6e, 0x0a, 0x53, 0x53, 0x68, 0x67, 0x6e, 0x44, 0x47, 0x7a, 0x39, 0x65,
  0x49, 0x47, 0x76, 0x55, 0x54, 0x6b, 0x39, 0x6c, 0x68, 0x30, 0x57, 0x31, 0x41, 0x64, 0x65, 0x76,
  0x68, 0x68, 0x37, 0x7a, 0x31, 0x5a, 0x64, 0x46, 0x4a, 0x70, 0x44, 0x6d, 0x58, 0x37, 0x6d, 0x34,
  0x33, 0x65, 0x73, 0x4d, 0x5a, 0x4a, 0x52, 0x4c, 0x4f, 0x61, 0x44, 0x4c, 0x4c, 0x51, 0x58, 0x61,
  0x74, 0x51, 0x55, 0x6f, 0x78, 0x4d, 0x0a, 0x67, 0x6a, 0x64, 0x6f, 0x2f, 0x79, 0x6c, 0x4d, 0x4b,
  0x63, 0x68, 0x41, 0x49, 0x46, 0x68, 0x54, 0x67, 0x41, 0x45, 0x66, 0x4d, 0x72, 0x4c, 0x50, 0x55,
  0x6b, 0x6b, 0x75, 0x4b, 0x47, 0x39, 0x79, 0x6d, 0x73, 0x57, 0x5a, 0x31, 0x6a, 0x51, 0x54, 0x33,
  0x51, 0x4b, 0x42, 0x67, 0x51, 0x44, 0x39, 0x75, 0x52, 0x63, 0x41, 0x39, 0x4b, 0x56, 0x45, 0x5a,
  0x73, 0x6f, 0x41, 0x33, 0x68, 0x37, 0x5a, 0x0a, 0x57, 0x75, 0x48, 0x54, 0x58, 0x5a, 0x49, 0x56,
  0x33, 0x4b, 0x4f, 0x2f, 0x57, 0x69, 0x4f, 0x78, 0x6a, 0x33, 0x6e, 0x33, 0x54, 0x66, 0x4c, 0x6b,
  0x68, 0x54, 0x57, 0x4e, 0x45, 0x44, 0x76, 0x75, 0x36, 0x63, 0x76, 0x5a, 0x69, 0x78, 0x30, 0x75,
  0x76, 0x48, 0x45, 0x77, 0x6e, 0x33, 0x74, 0x49, 0x6b, 0x67, 0x4e, 0x70, 0x46, 0x55, 0x4b, 0x2b,
  0x78, 0x51, 0x38, 0x71, 0x35, 0x4a, 0x71, 0x32, 0x0a, 0x41, 0x46, 0x52, 0x6e, 0x67, 0x61, 0x68,
  0x72, 0x74, 0x41, 0x59, 0x47, 0x54, 0x75, 0x6e, 0x4a, 0x6d, 0x4a, 0x56, 0x44, 0x4c, 0x32, 0x57,
  0x57, 0x4c, 0x70, 0x66, 0x45, 0x4f, 0x45, 0x6e, 0x74, 0x46, 0x42, 0x6e, 0x65, 0x45, 0x48, 0x5a,
  0x2b, 0x53, 0x44, 0x6e, 0x68, 0x4f, 0x30, 0x46, 0x4d, 0x57, 0x50, 0x6d, 0x53, 0x6d, 0x42, 0x36,
  0x5a, 0x48, 0x6d, 0x58, 0x4d, 0x76, 0x70, 0x56, 0x56, 0x0a, 0x4b, 0x31, 0x2f, 0x45, 0x35, 0x67,
  0x52, 0x43, 0x32, 0x45, 0x4a, 0x5a, 0x6d, 0x4d, 0x47, 0x4b, 0x43, 0x6b, 0x75, 0x49, 0x39, 0x32,
  0x75, 0x5a, 0x50, 0x77, 0x4b, 0x42, 0x67, 0x51, 0x44, 0x55, 0x43, 0x50, 0x4c, 0x32, 0x4f, 0x63,
  0x73, 0x4d, 0x45, 0x46, 0x34, 0x55, 0x79, 0x6c, 0x34, 0x74, 0x73, 0x59, 0x32, 0x39, 0x42, 0x37,
  0x34, 0x30, 0x6f, 0x52, 0x6e, 0x30, 0x2b, 0x56, 0x78, 0x58, 0x0a, 0x74, 0x56, 0x6c, 0x4d, 0x54,
  0x44, 0x63, 0x61, 0x4c, 0x64, 0x75, 0x32, 0x46, 0x5a, 0x74, 0x63, 0x55, 0x65, 0x4d, 0x44, 0x33,
  0x6e, 0x42, 0x65, 0x41, 0x68, 0x32, 0x37, 0x51, 0x33, 0x49, 0x62, 0x73, 0x32, 0x2f, 0x7a, 0x6e,
  0x68, 0x56, 0x52, 0x5a, 0x49, 0x30, 0x46, 0x68, 0x75, 0x68, 0x30, 0x2f, 0x42, 0x4e, 0x32, 0x4e,
  0x6b, 0x56, 0x47, 0x6a, 0x55, 0x69, 0x52, 0x64, 0x54, 0x73, 0x51, 0x0a, 0x51, 0x45, 0x6f, 0x42,
  0x53, 0x6e, 0x67, 0x49, 0x4d, 0x2b, 0x2f, 0x77, 0x71, 0x58, 0x5a, 0x43, 0x2f, 0x78, 0x35, 0x6f,
  0x64, 0x56, 0x6c, 0x76, 0x78, 0x73, 0x62, 0x43, 0x74, 0x41, 0x51, 0x6d, 0x46, 0x32, 0x31, 0x30,
  0x6f, 0x69, 0x64, 0x38, 0x2f, 0x78, 0x6b, 0x37, 0x71, 0x6d, 0x74, 0x39, 0x4d, 0x70, 0x31, 0x44,
  0x74, 0x42, 0x4f, 0x6b, 0x65, 0x4a, 0x49, 0x63, 0x6f, 0x69, 0x78, 0x43, 0x0a, 0x64, 0x68, 0x37,
  0x54, 0x53, 0x6c, 0x77, 0x33, 0x4e, 0x77, 0x4b, 0x42, 0x67, 0x51, 0x43, 0x6e, 0x76, 0x66, 0x61,
  0x33, 0x61, 0x77, 0x62, 0x59, 0x63, 0x68, 0x79, 0x61, 0x39, 0x6a, 0x37, 0x75, 0x39, 0x47, 0x62,
  0x59, 0x6c, 0x73, 0x51, 0x2b, 0x4c, 0x2b, 0x45, 0x53, 0x59, 0x55, 0x57, 0x33, 0x33, 0x6d, 0x4f,
  0x45, 0x58, 0x4d, 0x52, 0x34, 0x30, 0x4a, 0x6a, 0x57, 0x36, 0x69, 0x53, 0x31, 0x0a, 0x6a, 0x31,
  0x78, 0x32, 0x4e, 0x73, 0x4c, 0x55, 0x59, 0x30, 0x46, 0x79, 0x49, 0x2f, 0x34, 0x64, 0x76, 0x47,
  0x73, 0x45, 0x7a, 0x6b, 0x6d, 0x53, 0x56, 0x6a, 0x58, 0x77, 0x35, 0x62, 0x4a, 0x6a, 0x49, 0x56,
  0x53, 0x73, 0x47, 0x63, 0x78, 0x2f, 0x61, 0x45, 0x57, 0x54, 0x5a, 0x68, 0x69, 0x71, 0x62, 0x33,
  0x59, 0x2f, 0x4e, 0x4e, 0x61, 0x30, 0x73, 0x35, 0x6b, 0x69, 0x34, 0x52, 0x50, 0x2b, 0x0a, 0x36,
  0x59, 0x73, 0x77, 0x68, 0x54, 0x48, 0x52, 0x79, 0x32, 0x35, 0x34, 0x75, 0x66, 0x51, 0x78, 0x78,
  0x61, 0x45, 0x36, 0x69, 0x64, 0x65, 0x77, 0x34, 0x36, 0x66, 0x70, 0x39, 0x52, 0x5a, 0x46, 0x65,
  0x4b, 0x65, 0x59, 0x52, 0x58, 0x6f, 0x4c, 0x74, 0x32, 0x34, 0x6d, 0x54, 0x68, 0x65, 0x6d, 0x61,
  0x36, 0x67, 0x45, 0x70, 0x59, 0x38, 0x39, 0x67, 0x51, 0x4b, 0x42, 0x67, 0x46, 0x6e, 0x77, 0x0a,
  0x66, 0x64, 0x61, 0x78, 0x78, 0x77, 0x34, 0x54, 0x39, 0x30, 0x52, 0x4d, 0x58, 0x73, 0x78, 0x31,
  0x44, 0x78, 0x46, 0x4b, 0x73, 0x6f, 0x53, 0x79, 0x66, 0x55, 0x63, 0x72, 0x61, 0x74, 0x38, 0x4e,
  0x58, 0x70, 0x5a, 0x39, 0x6b, 0x62, 0x58, 0x36, 0x45, 0x38, 0x4e, 0x56, 0x77, 0x50, 0x50, 0x56,
  0x44, 0x44, 0x31, 0x38, 0x34, 0x68, 0x38, 0x4f, 0x4f, 0x4c, 0x78, 0x52, 0x4c, 0x4a, 0x72, 0x52,
  0x0a, 0x62, 0x30, 0x42, 0x67, 0x49, 0x74, 0x7a, 0x47, 0x6b, 0x51, 0x62, 0x79, 0x4a, 0x62, 0x4a,
  0x52, 0x70, 0x70, 0x4f, 0x7a, 0x43, 0x34, 0x6d, 0x55, 0x72, 0x6c, 0x35, 0x38, 0x34, 0x70, 0x50,
  0x50, 0x49, 0x56, 0x55, 0x51, 0x32, 0x48, 0x39, 0x71, 0x51, 0x48, 0x48, 0x4f, 0x41, 0x41, 0x73,
  0x46, 0x47, 0x6b, 0x6e, 0x6d, 0x31, 0x51, 0x77, 0x54, 0x76, 0x67, 0x57, 0x69, 0x78, 0x45, 0x76,
  0x41, 0x0a, 0x44, 0x6a, 0x36, 0x31, 0x39, 0x5a, 0x67, 0x51, 0x77, 0x51, 0x78, 0x4e, 0x43, 0x70,
  0x45, 0x74, 0x6f, 0x67, 0x57, 0x77, 0x34, 0x64, 0x41, 0x58, 0x45, 0x4b, 0x73, 0x79, 0x6f, 0x79,
  0x47, 0x42, 0x35, 0x79, 0x74, 0x71, 0x31, 0x46, 0x6a, 0x62, 0x41, 0x6f, 0x47, 0x41, 0x62, 0x39,
  0x35, 0x5a, 0x70, 0x42, 0x56, 0x61, 0x51, 0x5a, 0x75, 0x31, 0x54, 0x45, 0x72, 0x62, 0x73, 0x63,
  0x46, 0x76, 0x0a, 0x4d, 0x62, 0x2f, 0x65, 0x57, 0x70, 0x32, 0x49, 0x41, 0x49, 0x2b, 0x75, 0x5a,
  0x2f, 0x73, 0x75, 0x72, 0x49, 0x5a, 0x5a, 0x73, 0x70, 0x4a, 0x39, 0x35, 0x45, 0x6c, 0x32, 0x75,
  0x33, 0x79, 0x44, 0x4f, 0x48, 0x37, 0x71, 0x45, 0x55, 0x73, 0x59, 0x65, 0x6f, 0x77, 0x41, 0x79,
  0x66, 0x77, 0x31, 0x44, 0x65, 0x55, 0x53, 0x4f, 0x73, 0x43, 0x74, 0x62, 0x65, 0x47, 0x6c, 0x34,
  0x68, 0x70, 0x55, 0x0a, 0x43, 0x56, 0x6f, 0x76, 0x61, 0x4f, 0x73, 0x78, 0x50, 0x78, 0x43, 0x4a,
  0x4f, 0x2b, 0x4e, 0x74, 0x59, 0x64, 0x77, 0x68, 0x7a, 0x32, 0x32, 0x6a, 0x70, 0x79, 0x62, 0x6d,
  0x76, 0x2b, 0x53, 0x2f, 0x2b, 0x74, 0x76, 0x39, 0x46, 0x65, 0x6a, 0x6c, 0x69, 0x31, 0x62, 0x35,
  0x64, 0x5a, 0x6b, 0x41, 0x47, 0x4a, 0x45, 0x48, 0x63, 0x63, 0x70, 0x32, 0x59, 0x52, 0x35, 0x76,
  0x4a, 0x69, 0x33, 0x7a, 0x0a, 0x42, 0x2f, 0x33, 0x6f, 0x2b, 0x6e, 0x63, 0x75, 0x53, 0x2b, 0x2b,
  0x53, 0x70, 0x51, 0x4f, 0x4c, 0x69, 0x46, 0x70, 0x53, 0x4f, 0x42, 0x55, 0x3d, 0x0a, 0x2d, 0x2d,
  0x2d, 0x2d, 0x2d, 0x45, 0x4e, 0x44, 0x20, 0x50, 0x52, 0x49, 0x56, 0x41, 0x54, 0x45, 0x20, 0x4b,
  0x45, 0x59, 0x2d, 0x2d, 0x2d, 0x2d, 0x2d, 0x0a,
};

